import tonio.colored as tonio


def test_scope_cancel(run):
    enter = []
    exit = []

    async def _sleep(idx, t):
        enter.append(idx)
        await tonio.sleep(t)
        exit.append(idx)

    async def _run():
        async with tonio.scope() as scope:
            scope.spawn(_sleep(1, 0.1))
            scope.spawn(_sleep(2, 2))
            await tonio.sleep(0.2)
            scope.cancel()
        await tonio.sleep(2)

    run(_run())

    assert set(enter) == {1, 2}
    assert set(exit) == {1}



from functools import wraps

from ._runtime import run


def main(
    *coros,
    context: bool = False,
    signals: list[int] | None = None,
    threads: int | None = None,
    blocking_threadpool_size: int = 128,
    blocking_threadpool_idle_ttl: int = 30,
):
    if not coros:
        #: opts
        def deco(coro):
            @wraps(coro)
            def wrapper():
                run(
                    coro(),
                    context=context,
                    signals=signals,
                    threads=threads,
                    blocking_threadpool_size=blocking_threadpool_size,
                    blocking_threadpool_idle_ttl=blocking_threadpool_idle_ttl,
                )

            return wrapper

        return deco

    if len(coros) > 1:
        raise SyntaxError('Invalid argument for `main`')

    [coro] = coros

    @wraps(coro)
    def wrapper():
        run(coro())

    return wrapper



import contextlib
from typing import TypeVar

from ._events import Event
from ._tonio import CancelledError, ResultHolder, get_runtime
from ._types import Coro


_T = TypeVar('_T')


class _Interval:
    __slots__ = ['deadline', 'len']

    def __init__(self, deadline: int, len: int):
        self.deadline = deadline
        self.len = len

    def _poll(self):
        now = get_runtime()._clock
        if now >= self.deadline:
            next, delay = now + self.len, 0
        else:
            next = self.deadline + self.len
            delay = self.deadline - now
        self.deadline = next
        return delay


class Interval(_Interval):
    __slots__ = []

    def tick(self):
        timeout = self._poll()
        yield Event().waiter(timeout)


def time() -> float:
    return get_runtime()._clock / 1_000_000


def sleep(timeout: int | float) -> Coro[None]:
    yield from Event().wait(timeout)


def timeout(coro: Coro[_T], timeout: int | float) -> Coro[tuple[None | _T, bool]]:
    done = Event()
    res = ResultHolder()
    errs = []

    def wrapper():
        try:
            ret = yield coro
            res.store(ret)
        except CancelledError:
            pass
        except Exception as exc:
            errs.append(exc)
        finally:
            done.set()

    get_runtime()._spawn_pygen(wrapper())

    yield from done.wait(timeout)
    if not done.is_set():
        with contextlib.suppress(CancelledError):
            coro.throw(CancelledError)
        return None, False
    if errs:
        [err] = errs
        raise err
    return res.fetch(), True


def interval(period: int | float, at: int | None = None) -> Interval:
    period = round(max(0, period * 1_000_000))
    at = at or get_runtime()._clock
    return Interval(at, period)



import pytest

import tonio
import tonio._colored
import tonio._colored._net._socket
import tonio._colored._scope
import tonio._colored._time
import tonio._net._socket
import tonio._time
from tonio._runtime import Runtime
from tonio._utils import is_asyncg


_runtime = Runtime(threads=4, threads_blocking=8, threads_blocking_timeout=10, context=False, signals=[])
_runtime_wctx = Runtime(threads=4, threads_blocking=8, threads_blocking_timeout=10, context=True, signals=[])


@pytest.fixture(scope='function')
def runtime(monkeypatch):
    monkeypatch.setattr(tonio._ctl, 'get_runtime', lambda: _runtime)
    monkeypatch.setattr(tonio._scope, 'get_runtime', lambda: _runtime)
    monkeypatch.setattr(tonio._time, 'get_runtime', lambda: _runtime)
    monkeypatch.setattr(tonio._net._socket, 'get_runtime', lambda: _runtime)
    monkeypatch.setattr(tonio._colored._ctl, 'get_runtime', lambda: _runtime)
    monkeypatch.setattr(tonio._colored._scope, 'get_runtime', lambda: _runtime)
    monkeypatch.setattr(tonio._colored._time, 'get_runtime', lambda: _runtime)
    monkeypatch.setattr(tonio._colored._net._socket, 'get_runtime', lambda: _runtime)
    return _runtime


@pytest.fixture(scope='function')
def runtime_with_ctx(monkeypatch):
    monkeypatch.setattr(tonio._ctl, 'get_runtime', lambda: _runtime_wctx)
    monkeypatch.setattr(tonio._scope, 'get_runtime', lambda: _runtime)
    monkeypatch.setattr(tonio._time, 'get_runtime', lambda: _runtime_wctx)
    monkeypatch.setattr(tonio._net._socket, 'get_runtime', lambda: _runtime_wctx)
    monkeypatch.setattr(tonio._colored._ctl, 'get_runtime', lambda: _runtime_wctx)
    monkeypatch.setattr(tonio._colored._scope, 'get_runtime', lambda: _runtime)
    monkeypatch.setattr(tonio._colored._time, 'get_runtime', lambda: _runtime_wctx)
    monkeypatch.setattr(tonio._colored._net._socket, 'get_runtime', lambda: _runtime_wctx)
    return _runtime_wctx


@pytest.fixture(scope='function')
def run(runtime):
    def inner(coro):
        runner = runtime.run_pyasyncgen_until_complete if is_asyncg(coro) else runtime.run_pygen_until_complete
        return runner(coro)

    return inner


@pytest.fixture(scope='function')
def run_with_ctx(runtime_with_ctx):
    def inner(coro):
        runner = (
            runtime_with_ctx.run_pyasyncgen_until_complete
            if is_asyncg(coro)
            else runtime_with_ctx.run_pygen_until_complete
        )
        return runner(coro)

    return inner



from .._net._hilvl import (  # noqa: F401
    open_tls_over_tcp_listeners as open_ssl_over_tcp_listeners,
    open_tls_over_tcp_stream as open_ssl_over_tcp_stream,
    serve_tls_over_tcp as serve_ssl_over_tcp,
)
from .._net._tls import TLSListener as SSLListener, TLSStream as SSLStream  # noqa: F401



from types import TracebackType

from .._sync import _LockImpl, _SemaphoreImpl
from .._tonio import (
    Barrier as _Barrier,
    Channel as _Channel,
    ChannelReceiver as _ChannelReceiver,
    ChannelSender as _ChannelSender,
    UnboundedChannel as _UnboundedChannel,
    UnboundedChannelReceiver as _UnboundedChannelReceiver,
    UnboundedChannelSender as UnboundedChannelSender,
)


class Lock(_LockImpl):
    async def __aenter__(self):
        if event := self.acquire():
            await event.waiter(None)

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ):
        self.release()


class Semaphore(_SemaphoreImpl):
    async def __aenter__(self):
        if event := self.acquire():
            await event.waiter(None)

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ):
        self.release()


class Barrier(_Barrier):
    async def wait(self) -> int:
        count = self.ack()
        await self._event.waiter(None)
        return count


class ChannelSender(_ChannelSender):
    async def send(self, message) -> None:
        await self._send(message).waiter(None)


class ChannelReceiver(_ChannelReceiver):
    async def receive(self):
        while True:
            event, blocking, message = self._receive()
            if not blocking:
                return message
            await event.waiter(None)


class UnboundedChannelReceiver(_UnboundedChannelReceiver):
    async def receive(self):
        while True:
            event, blocking, message = self._receive()
            if not blocking:
                return message
            await event.waiter(None)


def channel(size) -> tuple[ChannelSender, ChannelReceiver]:
    inner = _Channel(size)
    return ChannelSender(inner), ChannelReceiver(inner)


def unbounded_channel() -> tuple[UnboundedChannelSender, UnboundedChannelReceiver]:
    inner = _UnboundedChannel()
    return UnboundedChannelSender(inner), UnboundedChannelReceiver(inner)



import contextlib
import errno
import signal
import threading

from ._tonio import CancelledError, get_runtime


class _SignalReceiver:
    __slots__ = ['_sigs', '_chr', '_chw', '_inner']

    def __init__(self, sigs):
        self._sigs = sigs

    def _init_channel(self):
        raise NotImplementedError

    def _register_coros(self, runtime):
        raise NotImplementedError

    def __enter__(self):
        self._chw, self._chr = self._init_channel()
        self._register_coros(get_runtime())
        return self

    def __exit__(self, exc_type, exc, tb):
        runtime = get_runtime()
        for sig in self._sigs:
            runtime._sig_rem(sig)
        with contextlib.suppress(CancelledError):
            self._inner.throw(CancelledError)

    def __iter__(self):
        return self

    def __aiter__(self):
        return self


def _noop(*args, **kwargs):
    return


def _is_main_thread():
    return threading.main_thread().ident == threading.current_thread().ident


def _check_sig(sig):
    if not isinstance(sig, int):
        raise TypeError(f'sig must be an int, not {sig!r}')

    if sig not in signal.valid_signals():
        raise ValueError(f'invalid signal number {sig}')


def _set_sig_wfd(fd):
    if fd >= 0:
        return signal.set_wakeup_fd(fd, warn_on_full_buffer=False)
    return signal.set_wakeup_fd(fd)


def _sig_add(sig):
    if not _is_main_thread():
        raise ValueError('Signals can only be handled from the main thread')

    _check_sig(sig)
    try:
        # register a dummy signal handler so Python will write the signal no in the wakeup fd
        signal.signal(sig, _noop)
        # set SA_RESTART to limit EINTR occurrences
        signal.siginterrupt(sig, False)
    except OSError as exc:
        if exc.errno == errno.EINVAL:
            raise RuntimeError(f'signum {sig} cannot be caught')
        raise


def _sig_rem(sig):
    if not _is_main_thread():
        raise ValueError('Signals can only be handled from the main thread')

    if sig == signal.SIGINT:
        handler = signal.default_int_handler
    else:
        handler = signal.SIG_DFL

    try:
        signal.signal(sig, handler)
    except OSError as exc:
        if exc.errno == errno.EINVAL:
            raise RuntimeError(f'signum {sig} cannot be caught')
        raise


def _signal_receiver(cls, *signals):
    for sig in signals:
        _check_sig(sig)
    return cls(signals)



import time

import tonio
import tonio.time


def test_time(run):
    stack = []

    def _run():
        stack.append(tonio.time.time())
        yield
        stack.append(tonio.time.time())

    run(_run())
    assert stack[1] > stack[0]


def test_sleep(run):
    def _run():
        start = time.monotonic()
        yield tonio.spawn(tonio.time.sleep(0.05), tonio.time.sleep(0.1))
        return time.monotonic() - start

    assert run(_run()) >= 0.1


def test_timeout(run):
    stack = []

    def _sleep(x):
        yield tonio.time.sleep(x)
        stack.append(x)
        return 3

    def _run():
        out1, success1 = yield tonio.time.timeout(_sleep(0.2), 0.3)
        out2, success2 = yield tonio.time.timeout(_sleep(0.2), 0.1)
        yield tonio.time.sleep(1)
        return (out1, out2, success1, success2)

    out1, out2, success1, success2 = run(_run())
    assert out1 == 3
    assert out2 is None
    assert success1 is True
    assert success2 is False
    assert len(stack) == 1


def test_interval(run):
    stack = []
    event = tonio.Event()

    def _scheduler(interval):
        times = 0
        t0 = tonio.time.time()
        while times < 5:
            yield interval.tick()
            t = tonio.time.time()
            stack.append((t, (t - t0) / (times or 1)))
            yield tonio.sleep(0.01)
            times += 1
        event.set()

    def _run():
        interval = tonio.time.interval(0.02)
        tonio.spawn(_scheduler(interval))
        yield event.wait()

    run(_run())
    assert len(stack) == 5
    assert all(v[1] >= 0.02 for v in stack[1:])



import ssl

import pytest
import trustme

import tonio
from tonio.net import socket
from tonio.net.ssl import SSLStream, open_ssl_over_tcp_stream, serve_ssl_over_tcp


_SIZE = 1024 * 1024


@pytest.fixture(scope='session')
def ssl_ca():
    return trustme.CA()


@pytest.fixture(scope='session')
def ssl_cert(ssl_ca):
    return ssl_ca.issue_server_cert('127.0.0.1')


@pytest.fixture(scope='function')
def ssl_server_ctx(ssl_cert):
    ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
    ssl_cert.configure_cert(ctx)
    return ctx


@pytest.fixture(scope='function')
def ssl_client_ctx(ssl_ca):
    ctx = ssl.create_default_context()
    ssl_ca.configure_trust(ctx)
    return ctx


def _get_port():
    sock = socket.socket()

    with sock:
        yield sock.bind(('127.0.0.1', 0))
        name = sock.getsockname()
        return name[1]


def test_ssl_tcp_recv(run, ssl_server_ctx, ssl_client_ctx):
    def server():
        done = tonio.Event()
        res = []
        port = yield _get_port()

        def _server_handler(stream: SSLStream):
            buf = b''
            while len(buf) < _SIZE:
                buf += yield stream.receive_some()
            res.append(buf)
            done.set()

        with tonio.scope() as scope:
            scope.spawn(serve_ssl_over_tcp(_server_handler, host='127.0.0.1', port=port, ssl_context=ssl_server_ctx))
            scope.spawn(client(port))
            yield done.wait()
            scope.cancel()
        yield scope()

        return res[0]

    def client(port):
        yield tonio.sleep(0.5)
        stream: SSLStream = yield open_ssl_over_tcp_stream('127.0.0.1', port=port, ssl_context=ssl_client_ctx)
        yield stream.send_all(b'a' * _SIZE)

    data = run(server())
    assert data == b'a' * _SIZE


def test_streams_tcp_send(run, ssl_server_ctx, ssl_client_ctx):
    done = tonio.Event()
    state = {'data': b''}

    def server():
        port = yield _get_port()

        def _server_handler(stream: SSLStream):
            yield stream.send_all(b'a' * _SIZE)

        with tonio.scope() as scope:
            scope.spawn(serve_ssl_over_tcp(_server_handler, host='127.0.0.1', port=port, ssl_context=ssl_server_ctx))
            scope.spawn(client(port))
            yield done.wait()
            scope.cancel()
        yield scope()

    def client(port):
        yield tonio.sleep(0.5)
        stream: SSLStream = yield open_ssl_over_tcp_stream('127.0.0.1', port=port, ssl_context=ssl_client_ctx)
        while len(state['data']) < _SIZE:
            state['data'] += yield stream.receive_some()
        done.set()

    run(server())
    assert state['data'] == b'a' * _SIZE



from ..._colored._net._hilvl import (
    open_tcp_listeners as open_tcp_listeners,
    open_tcp_stream as open_tcp_stream,
    open_unix_socket as open_unix_socket,
    serve_listeners as serve_listeners,
    serve_tcp as serve_tcp,
)
from ..._colored._net._streams import SocketListener as SocketListener, SocketStream as SocketStream
